#!/usr/bin/env python
#
# Copyright 2020, Julian Catchen <jcatchen@illinois.edu>
#
# This file is part of Chromonomer.
#
# Chromonomer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Chromonomer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import argparse
import gzip
from datetime import datetime

agp_path   = ""
fasta_path = ""

def parse_command_line():
    global agp_path, fasta_path

    p = argparse.ArgumentParser()

    #
    # Add options.
    #
    p.add_argument("--agp", type=str,
                   help="Path to write the generated AGP file [default: stdout].")
    p.add_argument("--fasta", type=str,
                   help="Path to the input FASTA file from which to generate the AGP file [default stdin].")

    #
    # Parse the command line
    #
    args = p.parse_args()

    if args.agp != None:
        agp_path = args.agp
    if args.fasta != None:
        fasta_path = args.fasta


def load_fasta_file(fasta_path, fasta, ordered_ids):

    if len(fasta_path) > 0:
        if fasta_path[-3:] == ".gz":
            fh = gzip.open(fasta_path, 'rb')
        else:
            fh = open(fasta_path)
    else:
        fh = sys.stdin

    #
    # Load a FASTA file into a dictionary.
    #
    eof  = False
    line = fh.readline()
    if len(line) == 0:
        return
    line = line.strip('\n ')

    while len(line) == 0 or line[0] == "#":
        line = fh.readline()
        if len(line) == 0:
            return
        line = line.strip('\n ')

    while eof == False:
        id  = ""
        seq = ""

        if line[0] == ">":
            id   = line[1:]
            # Remove any comments from the FASTA ID line.
            pos  = id.find(' ')
            if pos >= 0:
                id = id[0:pos]
            line = ""

        while len(line) == 0 or line[0] != ">":
            line = fh.readline()
            if len(line) == 0:
                eof = True
                break
            line = line.strip('\n ')
            if len(line) == 0 or line[0] == "#":
                continue
            if line[0] != ">":
                seq += line
        #
        # Record the final sequence.
        #
        if id in fasta:
            print >> sys.stderr, "Warning: FASTA ID '" + id + "' appears more than one time in the input file, keeping last record."
        fasta[id] = seq
        ordered_ids.append(id)

    fh.close()

    path = fasta_path if len(fasta_path) > 0 else "STDIN"
    
    print >> sys.stderr, "Parsed '" + path + "', loaded " + str(len(fasta)) + " sequences."


def parse_fasta_seqs(fasta, scaffolds):

    for id in fasta:
        seq      = fasta[id]
        seq_len  = len(seq)
        scaffold = []
        
        pos   = 0
        start = pos + 1
        end   = start
        while pos < seq_len:
            if seq[pos] == 'N' or seq[pos] == 'n':
                end = pos - 1
                scaffold.append( ("ctg", start, end) )

                gap_start = pos
                gap_end   = pos

                while seq[pos] == 'N' or seq[pos] == 'n':
                    pos += 1

                gap_end = pos - 1
                scaffold.append( ("gap", gap_start, gap_end) )

                start = pos
            pos += 1
        end = pos - 1
        scaffold.append( ("ctg", start, end) )
        scaffolds[id] = scaffold

    return
    
def write_agp_file(agp_path, ordered_ids, scaffolds):
    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    if len(agp_path) > 0:
        if agp_path[-3:] == ".gz":
            fh = gzip.open(agp_path, 'wb')
        else:
            fh = open(agp_path)
    else:
        fh = sys.stdout

    d = datetime.now()
    fh.write("# AGP file generated on " + d.strftime("%Y%m%d %H:%M:%S") + "\n" +
             "# Command: " + " ".join(sys.argv) + "\n#\n")
        
    ctg_cnt = 0
    for scaffold in scaffolds:
        for comp in scaffold:
            if comp == "ctg":
                ctg_cnt += 1

    if ctg_cnt < 100:
        width = '02'
    elif ctg_cnt < 1000:
        width = '03'
    elif ctg_cnt < 10000:
        width = '04'
    else:
        width = '05'
        
    ctg_id  = 1
    gap_cnt = 0
    for id in ordered_ids:
        scaffold = scaffolds[id]
        part_no  = 1

        for comp in scaffold:
            (type, start, end) = comp

            if type == "ctg":
                fh.write('{}\t{}\t{}\t{}\tW\tctg{:{}}\t1\t{}\t+\n'.format(id, start, end, part_no, ctg_id, width, end - start + 1))
                ctg_id  += 1
                ctg_cnt += 1
            else:
                fh.write('{}\t{}\t{}\t{}\tN\t{}\tfragment\tyes\t\n'.format(id, start, end, part_no, end - start + 1))
                gap_cnt += 1
            part_no += 1

    print >> sys.stderr, "Wrote AGP information for " + str(len(scaffolds)) + " scaffolds, consisting of " + str(ctg_cnt) + " contigs."

    fh.close()

#                                                                              #
#------------------------------------------------------------------------------#
#                                                                              #

fasta       = {}
ordered_ids = []
scaffolds   = {}

parse_command_line()

load_fasta_file(fasta_path, fasta, ordered_ids)

parse_fasta_seqs(fasta, scaffolds)

write_agp_file(agp_path, ordered_ids, scaffolds)
