#!/usr/bin/perl
#
# Copyright 2015, Julian Catchen <jcatchen@illinois.edu>
#

use strict;

use constant false => 0;
use constant true  => 1;

my $agp_path = "";
my $unp_path = "";
my $scaffold = "";
my $bp       = "";

my (%agp, %agp_index, %unplaced_scaffolds);

parse_command_line();

load_agp_file($agp_path, \%agp);

load_unplaced_scaffolds($unp_path, \%unplaced_scaffolds);

index_agp(\%agp, \%agp_index);


if (!defined($unplaced_scaffolds{$scaffold}) && !defined($agp_index{$scaffold})) {
    print STDERR "Unable to locate the scaffold specified, '$scaffold'\n";
    exit(1);

} elsif (defined($unplaced_scaffolds{$scaffold})) {
    print STDERR "The scaffold specificed ('$scaffold') was not integrated into the assembly.\n";
    exit(1);
}

#
# Find the proper fragment of this scaffold that this component falls within.
#
my ($new_chr, $new_bp, $i);
my $found = false;
for ($i = 0; $i < scalar(@{$agp_index{$scaffold}}); $i++) {

    if ($bp >= $agp_index{$scaffold}->[$i]->{'sbp'} &&
	$bp <= $agp_index{$scaffold}->[$i]->{'ebp'}) {

	$found   = true;
	$new_chr = $agp_index{$scaffold}->[$i]->{'chr'};
	$new_bp  = $agp_index{$scaffold}->[$i]->{'chr_sbp'} + ($bp - $agp_index{$scaffold}->[$i]->{'sbp'});
    }

    last if ($found);
}

if ($found == false) {
    print STDERR "ERROR: Unable to find agp component for '$scaffold', $bp\n";
} else {
    print "Translated: $scaffold, $bp => $new_chr, $new_bp\n";
}


sub index_agp {
    my ($agp, $index) = @_;

    my ($chr, $scaf, $href);

    foreach $chr (keys %{$agp}) {

	foreach $href (@{$agp->{$chr}}) {

	    next if ($href->{'type'} eq "N");

	    if (!defined($index->{$href->{'scaf'}})) {
		$index->{$href->{'scaf'}} = [];
	    }

	    push(@{$index->{$href->{'scaf'}}}, 
		 {'sbp'     => $href->{'scaf_sbp'},
		  'ebp'     => $href->{'scaf_ebp'},
		  'ori'     => $href->{'scaf_ori'},
		  'chr_sbp' => $href->{'start'},
		  'chr_ebp' => $href->{'end'},
		  'chr'     => $href->{'chr'}
		 });
	}
    }

    #
    # Sort all the scaffold components according to starting basepair.
    #
    foreach $scaf (keys %{$index}) {
	@{$index->{$scaf}} = sort {$a->{'sbp'} <=> $b->{'sbp'}} @{$index->{$scaf}};
    }

    # foreach $scaf (keys %{$index}) {
    # 	print STDERR $scaf, ":\n";
    # 	foreach $href (@{$index->{$scaf}}) {
    # 	    print STDERR "  ", $href->{'sbp'}, " - ", $href->{'ebp'}, "; Chr ", $href->{'chr'}, "\n";
    # 	}
    # }
}

sub load_agp_file {
    my ($agp_path, $agp) = @_;

    my ($fh, @parts, $line, $href);

    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    open($fh, "<" . $agp_path) or die("Unable to open AGP file '$agp_path' $!\n");

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	next if (substr($parts[0], 0, 1) eq "#");

	$href = {};
	$href->{'chr'}   = $parts[0];
	$href->{'start'} = $parts[1] + 0;
	$href->{'end'}   = $parts[2] + 0;
	$href->{'part'}  = $parts[3];
	$href->{'type'}  = $parts[4];
	$href->{'len'}   = $href->{'type'} eq "N" ? $parts[5] : $parts[7] - $parts[6] + 1;

	if ($href->{'type'} ne "N") {
	    $href->{'scaf'}     = $parts[5];
	    $href->{'scaf_sbp'} = $parts[6] + 0;
	    $href->{'scaf_ebp'} = $parts[7] + 0;
	    $href->{'scaf_ori'} = $parts[8];
	}

	push(@{$agp->{$href->{'chr'}}}, $href);
    }

    print STDERR "Loaded gap information for ", scalar(keys %{$agp}), " chromosomes.\n";

    close($fh);
}

sub load_unplaced_scaffolds {
    my ($unp_path, $unplaced) = @_;

    my ($fh, @parts, $line, $href);

    return if (length($unp_path) == 0);

    open($fh, "<" . $unp_path) or die("Unable to open unplaced scaffolds AGP file '$unp_path' $!\n");

    while ($line = <$fh>) {
        chomp $line;
        @parts = split(/\t/, $line);

	next if (substr($parts[0], 0, 1) eq "#");

	$unplaced->{$parts[0]}++;
    }

    print STDERR "Loaded information for ", scalar(keys %{$unplaced}), " scaffolds.\n";

    close($fh);
}

sub parse_command_line {
    while (@ARGV) {
	$_ = shift @ARGV;
	if    ($_ =~ /^-s$/) { $scaffold = shift @ARGV; }
	elsif ($_ =~ /^-b$/) { $bp       = shift @ARGV; }
	elsif ($_ =~ /^-a$/) { $agp_path = shift @ARGV; }
	elsif ($_ =~ /^-u$/) { $unp_path = shift @ARGV; }
	elsif ($_ =~ /^-h$/) { usage(); }
	else {
	    print STDERR "Unknown command line option: '$_'\n";
	    usage();
	}
    }

    if (length($agp_path) == 0) {
	print STDERR "You must specify the AGP file defining assembly gaps.\n";
	usage();
    }
}

sub usage {
    print STDERR <<EOQ; 
translate_bp.pl -s scaffold -b basepair -a agp_path [-u agp_path] [-d] [-h]
    s: scaffold to translate.
    b: basepair to translate.
    a: path to the AGP formatted file defining assembly gaps.
    u: path to the unplaced scaffolds AGP file.
    h: display this help message.
    d: turn on debug output.

EOQ

exit(0);
}
