#!/usr/bin/env python
#
# Copyright 2014-2020, Julian Catchen <jcatchen@illinois.edu>
#
# This file is part of Chromonomer.
#
# Chromonomer is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Chromonomer is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import argparse
import re
import gzip
from operator import attrgetter
from datetime import datetime

agp_path     = ""
unp_scf_path = ""
gtf_scf_path = ""
gtf_chr_path = ""
out_path     = ""
ann_type     = "gtf"


def parse_command_line():
    global agp_path, gtf_chr_path, agp_scf_path, gtf_scf_path, unp_scf_path
    global out_path, ann_type

    p = argparse.ArgumentParser()

    #
    # Add options.
    #
    p.add_argument("--agp", type=str,
                   help="Path to an AGP file that describes the chromosome-level genome.")
    p.add_argument("--scaffolds", type=str,
                   help="Path to a file that describes additional scaffolds not containined in a chromosome.")
    p.add_argument("--gtf", type=str,
                   help="Path to a GTF file that describes the gene locations on a scaffold-level genome. Genes will be tranlated to chromosome level.")
    p.add_argument("--gtf_chr", type=str,
                   help="Path to a GTF file that describes the gene locations on a chromosome-level genome. Will break annotations back down to scaffold level.")
    p.add_argument("--gff", type=str,
                   help="Path to a GFF file that describes the gene locations on a scaffold-level genome. Equivalent to --gtf but for GFF format. Genes will be translated to chromosome level")
    p.add_argument("--gff_chr", type=str,
                   help="Path to a GFF file that describes the gene locations on a chromosome-level genome. Equivalent to --gtf_chr but for GFF format. Will break annotations back down to scaffold level.")
    p.add_argument("--out", type=str,
                   help="Directory to write the translated GTF file.")
    p.add_argument("--parse_gff", action="store_true",
                   help="Parse GFF files instead of GTF files.")

    #
    # Parse the command line
    #
    args = p.parse_args()

    if args.out != None:
        out_path = args.out
    if args.agp != None:
        agp_path = args.agp
    if args.scaffolds != None:
        unp_scf_path = args.scaffolds
    if args.gtf != None:
        gtf_scf_path = args.gtf
    if args.gtf_chr != None:
        gtf_chr_path = args.gtf_chr
    if args.gff != None:
        gtf_scf_path = args.gff
        ann_type = "gff"
    if args.gff_chr != None:
        gtf_chr_path = args.gff_chr
        ann_type = "gff"
    if args.parse_gff == True:
        ann_type = "gff"

    if len(out_path) == 0 or (len(out_path) > 0 and os.path.exists(out_path) == False):
        print >> sys.stderr, "You must specify the path to write the translated GTF and the directory must exist."
        p.print_help()
        sys.exit()
    if out_path[-1:] != "/":
        out_path += "/"
        
    if len(agp_path) == 0 or (len(agp_path) > 0 and os.path.exists(agp_path) == False):
        print >> sys.stderr, "You must specify the path to the chromosome-level AGP file."
        p.print_help()
        sys.exit()
    if len(gtf_scf_path) == 0 and len(gtf_chr_path) == 0:
        print >> sys.stderr, "You must specify either a scaffold-level GTF file, or a chromosome-level GTF file."
        p.print_help()
        sys.exit()
    if len(gtf_scf_path) > 0 and len(gtf_chr_path) > 0:
        print >> sys.stderr, "You must specify either a scaffold-level GTF file, or a chromosome-level GTF file, not both."
        p.print_help()
        sys.exit()
    if len(gtf_scf_path) > 0 and os.path.exists(gtf_scf_path) == False:
        print >> sys.stderr, "Unable to find GTF file at specified path: '" + gtf_scf_path + "'."
        p.print_help()
        sys.exit()
    if len(gtf_chr_path) > 0 and os.path.exists(gtf_chr_path) == False:
        print >> sys.stderr, "Unable to find GTF file at specified path: '" + gtf_chr_path + "'."
        p.print_help()
        sys.exit()


def process_scaffold_gtf(gtf_path, out_path, index, unplaced):

    if gtf_path[-3:] == ".gz":
        in_fh = gzip.open(gtf_path, 'rb')
        ziped = True
    else:
        in_fh = open(gtf_path)
        ziped = False

    #
    # Pull the GTF filename off the path to construct the output filename.
    #
    start = gtf_path.rfind("/")
    start = start+1 if start >= 0 else 0
    end   = gtf_path.rfind(".")
    if gtf_path[end:] == ".gz":
        end = gtf_path.rfind(".", 0, end)
    gtf_file = gtf_path[start:end]

    if ann_type == "gtf":
	gtf_file += "-translated.gtf"
    else:
	gtf_file += "-translated.gff"

    if ziped:
        gtf_file += ".gz"
        out_fh = gzip.open(out_path + gtf_file, "w")
    else:
        out_fh = open(out_path + gtf_file, "w")

    eof = False

    d = datetime.now()
    out_fh.write("# Translation completed: " + d.strftime("%Y%m%d %H:%M:%S") + "\n" +
                 "# Command: " + " ".join(sys.argv) + "\n#\n")

    mult_scaf = 0
    not_found = 0
    not_trans = 0
    skipped   = 0
    trans     = 0
    error     = 0
    total     = 0
    gene_success   = 0
    gene_fail      = 0
    gene_not_found = 0
    gene_mult_lgs  = 0
    missing_scaffolds = {}
    skipped_types     = {}

    #
    # Store parent relationships between components.
    #
    parent_key = {}
    
    #
    # We need to read all the GTF components for a gene at one time. If all components
    # cannot be mapped, map none of them and report an error.
    #

    #
    # Get the first non-comment line.
    #
    line       = ""
    line_num   = 0
    components = []

    while True:
        line      = in_fh.readline()
        line_num += 1
	if line[0] == "#":
	    out_fh.write(line)
	else:
	    break

    #
    # Determine the gene ID for this set of components.
    #
    line  = line.strip('\n')
    parts = line.split('\t')
    gene_id      = find_gene_id(parent_key, parts[8])
    next_gene_id = gene_id
    components.append(parts)

    while True:
	#
	# Retrieve the remaining components for this gene.
	#
        while True:
            while True:
                #
                # Skip lines with comments.
                #
                line      = in_fh.readline()
                line_num += 1
                if len(line) == 0:
                    eof = True
                    break
                if line[0] != "#":
                    break
            if eof:
                break

	    line  = line.strip('\n')
	    parts = line.split('\t')

            if parts[2] == "region" or parts[2] == "contig" or parts[2] == "scaffold" or parts[2] == "biological_region":
                skipped += 1
                if parts[2] not in skipped_types:
                    skipped_types[parts[2]] = 1
                continue
            
	    next_gene_id = find_gene_id(parent_key, parts[8])
            if next_gene_id == gene_id:
	        components.append(parts)
	    else:
                break

	com_not_trans = 0
	com_not_found = 0
	com_trans_err = 0
	com_success   = 0
	fails         = []
	scaffolds     = {}
	lgs           = {}

	#
	# Process all components together.
	#
	for aref in components:
	    scaf = aref[0]
	    sbp  = int(aref[3])
	    ebp  = int(aref[4])

	    #
	    # Test that all components for this gene are on the same scaffold prior to translation.
	    #
            if scaf not in scaffolds:
                scaffolds[scaf] = 0
	    scaffolds[scaf] += 1

	    #
	    # Find this scaffold in our assembly, translate the coordinates.
	    #
	    if scaf not in unplaced and scaf not in index:
                if scaf not in missing_scaffolds:
                    missing_scaffolds[scaf] = 0
		missing_scaffolds[scaf] += 1
		com_not_found += 1
		fails.append(aref[2])
		continue
	    elif scaf in unplaced:
		com_not_trans += 1
		continue

	    #
	    # Find the proper fragment of this scaffold that this component falls within.
	    #
	    found = False
	    for i in range(0, len(index[scaf])):

		if (sbp >= index[scaf][i]['sbp'] and
                    sbp <  index[scaf][i]['ebp'] and
                    ebp >  index[scaf][i]['sbp'] and
                    ebp <= index[scaf][i]['ebp']):

		    found = True

		    if (index[scaf][i]['ori'] == "+"):
                        # 
                        # Translate chromosome/basepair coordinates.
                        #
		        aref[0] = index[scaf][i]['chr']
		        aref[3] = str(index[scaf][i]['chr_sbp'] + (sbp - index[scaf][i]['sbp']))
		        aref[4] = str(index[scaf][i]['chr_sbp'] + (ebp - index[scaf][i]['sbp']))
                    else:
                        #
		        # If this scaffold is in a negative orientation, reverse the basepair coordinates.
		        #
                        trans_sbp = index[scaf][i]['sbp'] + (index[scaf][i]['ebp'] - ebp)
                        trans_ebp = index[scaf][i]['sbp'] + (index[scaf][i]['ebp'] - sbp)
		        aref[0]   = index[scaf][i]['chr']
		        aref[3]   = str(index[scaf][i]['chr_sbp'] + (trans_sbp - index[scaf][i]['sbp']))
		        aref[4]   = str(index[scaf][i]['chr_sbp'] + (trans_ebp - index[scaf][i]['sbp']))
                        aref[6]   = "-" if aref[6] == "+" else "+"

		    com_success += 1

		    #
		    # Test whether components span chromosomes.
		    #
                    if index[scaf][i]['chr'] not in lgs:
                        lgs[index[scaf][i]['chr']] = 0
		    lgs[index[scaf][i]['chr']] += 1

		    # print >> sys.stderr, "Translating " + scaf + " to " + index[scaf][i]['chr'] + "; " + str(sbp) + " to " + str(aref[3]) + " ; " + str(ebp) + " to " + str(aref[4]) + "; " + aref[8]

	        elif (sbp >= index[scaf][i]['sbp'] and
		      sbp <  index[scaf][i]['ebp'] and
		      ebp >  index[scaf][i]['sbp'] and
		      ebp >  index[scaf][i]['ebp']):
		    found          = True;
		    com_trans_err += 1
		    fails.append(aref[2])

		    # print >> sys.stderr, 
		    # 	"Failed to translate: $scaf to index[scaf][i]['chr']; ",
		    # 	"ann start: $sbp; chr start: index[scaf][i]['sbp']; ",
		    # 	"ann end: $ebp; chr end: index[scaf][i]['ebp']\n";

		elif (sbp <  index[scaf][i]['sbp'] and
		      sbp <  index[scaf][i]['ebp'] and
		      ebp >  index[scaf][i]['sbp'] and
		      ebp <= index[scaf][i]['ebp']):
		    found  = True;
		    com_trans_err += 1
		    fails.append(aref[2])

		    # print >> sys.stderr, 
		    # 	"Failed to translate: $scaf to index[scaf][i]->['chr']; ",
		    # 	"ann start: $sbp; chr start: index[scaf][i]->['sbp']; ",
		    # 	"ann end: $ebp; chr end: index[scaf][i]->['ebp']\n";

		if found:
                    break

	    if found == False:
		print >> sys.stderr, "ERROR: Unable to find component " + ",".join(aref)

	total     += len(components)
	not_trans += com_not_trans
	not_found += com_not_found
	error     += com_trans_err
	trans     += com_success
        if len(scaffolds) > 1:
	    mult_scaf += 1 

	if com_success == len(components):
	    #
	    # Output all the components;
	    #
            for aref in components:
		out_fh.write("\t".join(aref) + "\n")
	    gene_success += 1

	elif com_not_trans == len(components):
	    #
	    # Output all the components;
	    #
	    for aref in components:
		out_fh.write("\t".join(aref) + "\n")
	    gene_success += 1

	elif com_not_found == len(components):
	    gene_not_found += 1

        else:
	    print >> sys.stderr, (
                "Unable to translate " + gene_id + "; " + str(com_success) + " out of " + str(len(components)) +
	        " components succeeded. Fails included: " + ", ".join(fails)),
	    if len(lgs) > 1:
		gene_mult_lgs += 1
		print >> sys.stderr, "; spans multiple chrs: ", ", ".join(lgs.keys()),

	    print >> sys.stderr, " [last line " + str(line_num) + "]"
	    gene_fail += 1

	#
	# Prepare for the next component.
	#
	components = [];
	components.append(parts)
	gene_id = next_gene_id

        if eof:
            break

    in_fh.close()
    out_fh.close()

    if len(missing_scaffolds) > 0:
        print >> sys.stderr, "Scaffolds in the GFF/GTF file not found in the assembly: " + ", ".join(missing_scaffolds) + "\n"

    print >> sys.stderr, (
	str(mult_scaf) +" genes occurred on multiple scaffolds prior to translation.\n" +
        "Skipped " + str(skipped) + " components, including types: " + ", ".join(skipped_types.keys()) + "\n" +
	"Processed " + str(gene_success+gene_fail) + " genes; " + str(gene_success) + " translated, " + str(gene_fail) + " failures (" + str(gene_mult_lgs) + " on multiple chrs); " +
	str(gene_not_found) +" genes on mising scaffolds\n" +
	"Processed " + str(total) +" components; " +
	str(trans) + " translations; "  +str(not_trans) + " not translated; "+ str(error) + " translation errors; " +
	str(not_trans) + " components on unassembled scaffolds; " +
	str(not_found) + "  components in " + str(len(missing_scaffolds)) + " missing scaffolds.")

def process_chromosome_gtf(gtf_path, out_path, agp, unplaced):

    if gtf_path[-3:] == ".gz":
        in_fh = gzip.open(gtf_path, 'rb')
        ziped = true
    else:
        in_fh = open(gtf_path)
        ziped = false

    #
    # Pull the GTF filename off the path.
    #
    start = gtf_path.rfind("/")
    start = start+1 if start >= 0 else 0
    end   = gtf_path.rfind(".")
    if gtf_path[end:] == ".gz":
        end = gtf_path.rfind(".", 0, end)
    gtf_file = gtf_path[start:end]

    if ann_type == "gtf":
	gtf_file += "-translated.gtf"
    else:
	gtf_file += "-translated.gff"

    if ziped:
        gtf_file += ".gz"
        out_fh = gzip.open(out_path + gtf_file, 'w')
    else:    
        out_fh = open(out_path + gtf_file, "w")

    eof = False

    d = datetime.now()
    out_fh.write("# Translation completed: " + d.strftime("%Y%m%d %H:%M:%S") + "\n" +
                 "# Command: " + " ".join(sys.argv) + "\n#\n")
    
    mult_scaf = 0
    not_found = 0
    not_trans = 0
    trans     = 0
    error     = 0
    total     = 0
    gene_success   = 0
    gene_fail      = 0
    gene_not_found = 0
    gene_not_trans = 0

    #
    # We need to read all the GTF components for a gene at one time. If all components
    # cannot be mapped, map none of them and report an error.
    #

    #
    # Get the first non-comment line.
    #
    line       = ""
    components = []

    while True:
        line = in_fh.readline()
	if line[0] == "#":
	    out_fh.write(line)
	else:
	    break

    #
    # Determine the gene ID for this set of components.
    #
    line  = line.strip('\n')
    parts = line.split('\t')
    gene_id      = find_gene_id(parent_key, parts[8])
    next_gene_id = gene_id
    components.append(parts)

    while True:
	#
	# Retrieve the remaining components for this gene.
	#
        while True:
            while True:
                #
                # Skip lines with comments.
                #
                line = in_fh.readline()
                if len(line) == 0:
                    eof = True
                    break
                if line[0] != "#":
                    break
            if eof:
                break

	    line  = line.strip('\n')
	    parts = line.split('\t')
	    next_gene_id = find_gene_id(parent_key, parts[8])
            if next_gene_id == gene_id:
	        components.append(parts)
	    else:
                break

	com_not_trans = 0
        com_not_found = 0
	com_trans_err = 0
        com_skipped   = 0
        com_spans_scf = 0
	com_success   = 0
	fails         = []
        skipped_types = {}
        scaffolds     = {}
        
        #
        # Find the scaffold that this gene occurs within.
        #
        found    = False
        chr      = components[0][0]
	chr_sbp  = int(components[0][3])
	chr_ebp  = int(components[0][4])
        scaf     = ""
        scaf_sbp = 0
        scaf_ebp = 0
        scaf_ori = "+"

        if chr in unplaced:
            com_not_trans += len(components)

        else:
            for comp in agp[chr]:
	        if (chr_sbp >= comp['start'] and
                    chr_sbp <  comp['end'] and
                    chr_ebp >  comp['start'] and
                    chr_ebp <= comp['end']):
                    found    = True
                    scaf     = comp['scaf']
                    comp_sbp = comp['start']    # Starting position of the scaffold within the chromosome.
                    comp_ebp = comp['end']      # Ending position of the scaffold within the chromosome.
                    scaf_sbp = comp['scaf_sbp'] # Starting position of the unincorporated scaffold.
                    scaf_ebp = comp['scaf_ebp'] # Ending position of the unincorporated scaffold.
                    scaf_ori = comp['scaf_ori']
                    
                elif (chr_sbp <  comp['start'] and
                      chr_ebp >  comp['start'] and
                      chr_ebp <= comp['end']):
                    # Gene spans scaffolds
                    scaffolds[comp['scaf']] = 1
                
                elif (chr_sbp >= comp['start'] and
                      chr_sbp <  comp['end'] and
                      chr_ebp >  comp['end']):
                    # Gene spans scaffolds
                    scaffolds[comp['scaf']] = 1

            if len(scaffolds) > 1:
                com_spans_scf += len(components)

	    #
	    # Process all components together.
	    #
            if found:
	        for aref in components:
	            chr = aref[0]
	            sbp = int(aref[3])
	            ebp = int(aref[4])
               
		    # #
		    # # If this scaffold is in a negative orientation, reverse the basepair coordinates.
		    # #

	            if (sbp < chr_sbp or ebp > chr_ebp):
	                com_trans_err += 1
	                fails.append(aref[2])
                    else:
                        #
                        # Translate chromosome/basepair coordinates.
                        #
		        aref[0] = scaf
		        aref[3] = str( (scaf_sbp - 1) + (sbp - comp_sbp + 1) )
		        aref[4] = str( (scaf_sbp - 1) + (ebp - comp_sbp + 1) )
		        com_success += 1
            else:
                for aref in components:
                    fails.append(aref[2])
                    com_not_found += 1

	total     += len(components)
	not_trans += com_not_trans
        not_found += com_not_found
	error     += com_trans_err
	trans     += com_success

	if com_success == len(components):
	    #
	    # Output all the components;
	    #
            for aref in components:
		out_fh.write("\t".join(aref) + "\n")
	    gene_success += 1

	elif com_not_trans == len(components):
	    #
	    # Output all the components;
	    #
	    for aref in components:
		out_fh.write("\t".join(aref) + "\n")
	    gene_success += 1
            gene_not_trans += 1

        elif com_spans_scf == len(components):
            print >> sys.stderr, "Unable to translate " + gene_id + ", spans scaffolds " + ", ".join(scaffolds.keys())
            gene_fail += 1

        else:
	    print >> sys.stderr, (
                "Unable to translate " + gene_id + "; " + str(com_success) + " out of " + str(len(components)) +
	        " components succeeded. Fails included: " + ", ".join(fails))
	    gene_fail += 1

	#
	# Prepare for the next component.
	#
	components = [];
	components.append(parts)
	gene_id = next_gene_id

        if eof:
            break

    in_fh.close()
    out_fh.close()

    if len(missing_scaffolds) > 0:
        print >> sys.stderr, "\nScaffolds in the GFF/GTF file not found in the assembly: " + ", ".join(missing_scaffolds)

    print >> sys.stderr, (
	"Processed " + str(gene_success+gene_fail) + " genes; " + str(gene_success) + " translated, " + str(gene_fail) + " failures " +
	str(gene_not_trans) +" genes on unassembled scaffolds and not translated\n" +
	"Processed " + str(total) +" components; " +
	str(trans) + " translations; " + str(error) + " translation errors; " +
	str(not_trans) + " components on unassembled scaffolds; " + str(not_found) + " components not found in genome.")

def find_gene_id(parent_key, comment):
    comp_id   = ""
    parent_id = ""

    parts = comment.split(';')

    for s in parts:
	if ann_type == "gtf":
            match = re.match("^\s*gene_id \"?(.+)\"?\s*$", s)
	    if match:
                comp_id = match.group(1)
		return comp_id
        else:
            match = re.match("^\s*ID=\"?(.+)\"?\s*$", s)
	    if match:
                comp_id   = match.group(1)

            match = re.match("^\s*Parent=\"?(.+)\"?\s*$", s)
	    if match:
                parent_id = match.group(1)

    #
    # Add a new parent relationship.
    #
    if len(comp_id) > 0 and len(parent_id) > 0:
        parent_key[comp_id] = parent_id

    #
    # If a parent ID was given, try to convert the comp_id to the parent ID.
    #
    if len(parent_id) > 0:
        comp_id = parent_id
        while True:
            if comp_id in parent_key:
                comp_id = parent_key[comp_id]
            else:
                break

    return comp_id

def index_agp(agp, index):
    for chr in agp:
	for comp in agp[chr]:
            if comp['type'] != "W":
                continue

	    if comp['scaf'] not in index:
		index[comp['scaf']] = []

            index[comp['scaf']].append({
                'sbp':     comp['scaf_sbp'],
                'ebp':     comp['scaf_ebp'],
                'ori':     comp['scaf_ori'],
                'chr_sbp': comp['start'],
                'chr_ebp': comp['end'],
                'chr':     comp['chr']})

    #
    # Sort all the scaffold components according to starting basepair.
    #
    for scaf in index:
	index[scaf].sort(key=lambda a: a['sbp'])

    # foreach $scaf (keys %{$index}) {
    # 	print >> sys.stderr, $scaf, ":\n";
    # 	foreach $comp (@{$index->{$scaf}}) {
    # 	    print >> sys.stderr, "  ", $comp->['sbp'], " - ", $comp->['ebp'], "; Chr ", $comp->['chr'], "\n";
    # 	}
    # }


def load_agp_file(agp_path, agp):
    #
    # Load an AGP file defining gaps in the assembly.
    #  Format: object     object_beg object_end part_number component_type component_id component_beg component_end  orientation
    #  Gaps:   object     object_beg object_end part_number N              gap_length   gap_type      linkage        -empty-
    #          JH556661.1 1          70736 1 W AGAJ01000001.1 1        70736    +
    #          JH556661.1 70737      70782 2 N 46             fragment yes
    #
    if agp_path[-3:] == ".gz":
        fh = gzip.open(agp_path, 'rb')
    else:
        fh = open(agp_path)

    for line in fh:
        line  = line.strip('\n')
        parts = line.split('\t')

	if line[0] == "#":
             continue

        href = {}
	href['chr']   = parts[0]
	href['start'] = int(parts[1])
	href['end']   = int(parts[2])
	href['part']  = parts[3]
	href['type']  = parts[4]
	href['len']   = parts[5] 

	if href['type'] == "W":
            href['len']      = int(parts[7]) - int(parts[6]) + 1
	    href['scaf']     = parts[5]
	    href['scaf_sbp'] = int(parts[6])
	    href['scaf_ebp'] = int(parts[7])
	    href['scaf_ori'] = parts[8]

        if href['chr'] not in agp:
            agp[href['chr']] = []
	agp[href['chr']].append(href)

    print >> sys.stderr, "Loaded assembly information for " + str(len(agp)) + " chromosomes."

    fh.close()

def load_unplaced_scaffolds(unp_path, unplaced):
    if unp_path[-3:] == ".gz":
        fh = gzip.open(unp_path, 'rb')
    else:
        fh = open(unp_path)

    for line in fh:
        line  = line.strip('\n')
        parts = line.split('\t')

	if line[0] == "#":
            continue

        if parts[0] not in unplaced:
            unplaced[parts[0]] = 0
	unplaced[parts[0]] += 1

    print >> sys.stderr, "Loaded assembly information for " + str(len(unplaced)) + " additional unassembled scaffolds."

    fh.close()


#                                                                              #
#------------------------------------------------------------------------------#
#                                                                              #

agp                = {}
agp_index          = {}
unplaced_scaffolds = {}

parse_command_line()

load_agp_file(agp_path, agp)

load_unplaced_scaffolds(unp_scf_path, unplaced_scaffolds)

index_agp(agp, agp_index)

if len(gtf_scf_path) > 0:
    process_scaffold_gtf(gtf_scf_path, out_path, agp_index, unplaced_scaffolds)

if len(gtf_chr_path) > 0:
    process_chromosome_gtf(gtf_chr_path, out_path, agp, unplaced_scaffolds)
