// -*-mode:c++; c-style:k&r; c-basic-offset:4;-*-
//
// Copyright 2016-2019, Julian Catchen <jcatchen@illinois.edu>
//
// This file is part of Chromonomer.
//
// Chromonomer is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Chromonomer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Chromonomer.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef __MARKER_H__
#define __MARKER_H__

#include <utility>
using std::pair;
using std::make_pair;
#include <string>
using std::string;
#include <vector>
using std::vector;

#include "constants.h"
#include "Seq.h"

class Marker {
public:
    string      id;
    //
    // Alignment properties.
    //
    string      scaffold;
    uint        bp;
    strand_type strand;
    string      cigar;
    uint        aln_cnt;

    //
    // Unit-free ordering of this marker on its linkage group/scaffold.
    //
    size_t      lg_index;
    size_t      sc_index;

    //
    // Genetic map properties.
    //
    string      lg;
    double      cM;

    //
    // Record previous scaffold/basepairs if marker was relocated.
    //
    vector<pair<string, int> > old;

    Marker() {
	this->bp = 0;
	this->cM = 0.0;
        this->aln_cnt = 0;
    }
    Marker(string id) {
	this->id = id;
	this->bp = 0;
	this->cM = 0.0;
        this->aln_cnt = 0;
    }
    Marker(string id, string lg, double cM) {
	this->id = id;
	this->lg = lg;
	this->cM = cM;
	this->bp = 0;
        this->aln_cnt = 0;
    }
};

//
// Record clusters of consistent markers between the linkage map and scaffold.
//
class Cluster {
public:
    size_t   id;
    size_t   sliding_window;
    ori_type orientation;
    size_t   lg_start_index;
    size_t   lg_last_index;
    size_t   sc_start_index;
    size_t   sc_last_index;
    double   lg_start_cM;
    double   lg_end_cM;
    size_t   sc_start_bp;
    size_t   sc_end_bp;
    
    vector<Marker> members;

    Cluster() {
        this->orientation    = ori_forward;
	this->sliding_window = 0;
	this->lg_start_index = 0;
	this->lg_last_index  = 0;
        this->lg_start_cM    = 0.0;
        this->lg_end_cM      = 0.0;
	this->sc_start_index = 0;
	this->sc_last_index  = 0;
        this->sc_start_bp    = 0;
        this->sc_end_bp      = 0;
    }
    Cluster(ori_type ori, int sliding_window, size_t lg_start, size_t sc_start) {
        this->orientation    = ori;
	this->sliding_window = sliding_window;
	this->lg_start_index = lg_start;
	this->lg_last_index  = lg_start;
	this->sc_start_index = sc_start;
	this->sc_last_index  = sc_start;
    }
    void set(ori_type ori, int sliding_window, size_t lg_start, size_t sc_start) {
        this->orientation    = ori;
	this->sliding_window = sliding_window;
	this->lg_start_index = lg_start;
	this->lg_last_index  = lg_start;
	this->sc_start_index = sc_start;
	this->sc_last_index  = sc_start;
    }
    void reset() {
        this->members.clear();
        this->orientation    = ori_forward;
	this->sliding_window = 0;
	this->lg_start_index = 0;
	this->lg_last_index  = 0;
        this->lg_start_cM    = 0.0;
        this->lg_end_cM      = 0.0;
	this->sc_start_index = 0;
	this->sc_last_index  = 0;
        this->sc_start_bp    = 0;
        this->sc_end_bp      = 0;
    }
    friend ostream& operator<< (ostream& os, const Cluster &c) {
        os << "LG: " << c.lg_start_index << " - " << c.lg_last_index << " [" << c.lg_start_cM << " - " << c.lg_end_cM << "cM]; "
           << "Scaffold: " << c.sc_start_index << " - " << c.sc_last_index << " [" << c.sc_start_bp << " - " << c.sc_end_bp << "bp]\n"
           << "  Marker count: " << c.members.size() << "; "
           << "orientation: " << (c.orientation == ori_forward ? "forward" : "reverse") << "\n";
        return os;
    }

};

#endif // __MARKER_H__
